
from sqlalchemy import and_
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy import func,text
from sqlalchemy.future import select
from typing import TypeVar,Tuple,List,Dict,Any,Union,Iterator

from sqlalchemy.orm import QueryPropertyDescriptor,joinedload,subqueryload,contains_eager
from sqlalchemy.orm.query import Query
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.engine.row import  Row

from  sqlalchemy.engine.result import ScalarResult,ChunkedIteratorResult
from sqlalchemy.sql.elements import ColumnElement
from sqlalchemy.sql import Select
from .db_filter import absFilterItems,Page_param
from .po import BasePO
from .db_utils import db_tools
 
#from sqlalchemy.orm import selectinload # 紧急装载器 在该表主键又外键的基础上使用 select(UserTable).options(selectinload(UserTable.LoginLog))
 
from co6co.utils import log 


class DbOperations:
	# 实体类存在字段
	"""
	一些用法
	https://www.cnblogs.com/gcxblogs/p/14979274.html#subquery
	"""

	def __init__(self,db_session:AsyncSession) -> None:
		self.db_session = db_session 
		pass
	def _createQuery(self,poType:TypeVar):
		"""
		对没有 query 的 PO 增加 query 对象
		"""
		if not  hasattr(poType,"query") or isinstance(poType.query,Query): 
			query=self.db_session.query_property()  
			poType.query=query  
		 
	
	async def _get_one(self, select:select, selectField:bool=True):
		"""
		获取一行数据
		存在多,返回第一行
		为None 返回None
		存在多列：返回 Row
		""" 
		try: 
			data=await self.db_session.execute(select)
			if selectField:
				#row=data.one_or_none() # 没有返回None， 多行 异常Multiple rows were found when one or none was required
				row=data.fetchone() 	# 没有返回None， 多行，选一行给你
				if row==None:return row 
				return dict(zip(row._fields,row)) 
			else:
				row:Row=data.fetchone()   
				if row==None:return row 
				if len(row)==1:return row[0]
				else:return row # 返回后面这个不好使用 [{row._fields[i]:row[i]} for i in range(0,len(row))]
		except Exception as e:
			log.warn(f"在未找到!\terror:{e}")
			return None  
		
	async def _get_scalar(self ,select:select)->Any:
		data=await self.db_session.execute(select)
		return data.scalar()
	 
	async def _get_tuple(self ,select:select)-> List[dict]:  
		#sqlalchemy.engine.result.ChunkedIteratorResult
		data=await self.db_session.execute(select) 
		return [dict(zip(a._fields,a))  for a in  data]

	async def _get_list(self, select:select,remove_instance_state:bool=True)-> List[dict]|List[TypeVar]:  
		data=await self.db_session.execute(select) 
		if remove_instance_state:
			return db_tools.remove_db_instance_state(data.scalars().fetchall())  
		else:
			return data.scalars().fetchall() 
	
	async def get_one(self,selectColumnOrPo:TypeVar|Tuple[InstrumentedAttribute],*filters:ColumnElement[bool])->Any|None:
		"""
		获取一条数据
		返回 dict 或者 PO 实体
		"""
		'''result = await self.db_session.execute(text('select version()')) #通过conn对象就可对DB进行操作​ 
		data=result.fetchone() '''
		isTule,sml=self.create_select(selectColumnOrPo,*filters)
		return  await self._get_one(sml,isTule) 

	async def get_one_by_pk(self,po:TypeVar,pk:Union[Any, Tuple[Any, ...]]):
		"""
		通过 主键 获取实体
		"""
		try:
			one=await self.db_session.get_one(po,ident=pk)
			return one
		except Exception as e:
			log.warn(f"在{po},查找：pk:{pk} 未找到!\terror:{e}")
			return None
	 
	def create_select(self,selectColumnOrPo:Tuple[InstrumentedAttribute]|TypeVar,*filters:ColumnElement[bool] ) ->  Tuple[bool,Select[Any]]  :
		isTule= type(selectColumnOrPo)==tuple or type(selectColumnOrPo)==list  
		if isTule:
			sml=select(*selectColumnOrPo)  .filter(and_(*filters))
		else:
			# {type(selectColumnOrPo)} == sqlalchemy.orm.decl_api.DeclarativeMeta 
			po:TypeVar=selectColumnOrPo
			sml=select(po).filter(and_(*filters))
		return isTule,sml 
	
	def join(self,select:Select[Any], tarGet:TypeVar ,*filters:ColumnElement[bool] ) ->   Select[Any]:
		"""
		join , 没进行充分测试 谨慎使用
		"""
		sml=select.join(tarGet,and_(*filters)) 
		return sml
	

	async def _create_paged_select(self,filterItem:absFilterItems, selectColumnOrPo:Tuple[InstrumentedAttribute]|TypeVar=None,  allow_paged:bool=True, allow_order:bool=True)-> Tuple[bool,select] :
		"""
		创建 Column select
		"""
		filters=filterItem.filter()  
		#* selectColumn 不会为 None
		if selectColumnOrPo==None:selectColumnOrPo=filterItem.po_type 
		isTule,sml = self.create_select(selectColumnOrPo,*filters)  
		
		sml:Select[Any]=sml 
		if allow_paged:
			limit=filterItem.limit
			offset=filterItem.offset
			sml=sml.offset(offset).limit(limit)
		if allow_order:
			orderBy=filterItem.getOrderBy()
			sml=sml.order_by(*orderBy)
		return isTule,sml
	
	async def text(self,text:text)->dict:
		"""
		执行文本sql,
		返回: dict
		"""
		data=await self.db_session.execute(select(text))  
		return [dict(zip(a._fields,a))  for a in  data] 
	
	async def get_list(self,selectColumnOrPo:Tuple[InstrumentedAttribute]|TypeVar, *filters:ColumnElement[bool],remove_instance_state:bool=True): 
		isTule,sml=self.create_select(selectColumnOrPo,*filters)
		if isTule:return await self._get_tuple(sml)
		return await self._get_list(sml,remove_instance_state)
	
	async def count(self,*filters:ColumnElement[bool],column:InstrumentedAttribute="*" )->int:
		return await self._get_scalar(select(func.count(column)).filter(and_(*filters))) 
	
	async def exist(self,*filters:ColumnElement[bool],column:InstrumentedAttribute="*" )->bool:
		return await self.count(*filters,column=column)>0
	
	def add(self,instance: object, _warn: bool = True):
		self.db_session.add(instance,_warn)
	def add_all(self,instances: Iterator[object]):
		self.db_session.add_all(instances)

	async def delete(self,instance: object):
		await self.db_session.delete(instance)

	async def commit(self):
		await self.db_session.commit()
	async def rollback(self):
		await self.db_session.rollback()
	async def close(self):
		await self.db_session.close()

	
	 

    
	async def query_joined(self,poType:TypeVar, joinArr:Tuple[InstrumentedAttribute], *filters:ColumnElement[bool],param:Page_param=None,orderby: List[InstrumentedAttribute]=None):
		"""
		joinedload ==查询==> LEFT OUTER JOIN 
		不应该使用
		"""  
		self._createQuery(poType) 
		query=poType.query
		if len (joinArr)>0:
			query.options(joinedload(*joinArr)) 
		if  len (filters)>0:			
			query.filter(and_(*filters))
		if param !=None: 
			limit=param.limit
			offset=param.offset
			query.limit(limit).offset(offset)
		return (query.all()) 
	async def query_subquery(self,poType:TypeVar, joinArr:Tuple[InstrumentedAttribute], *filters:ColumnElement[bool],param:Page_param=None,orderby: List[InstrumentedAttribute]=None):
		"""
		先查询 users, 在 select * from (user) u join address on u.u_id == address.user_id
  		不应该使用
		"""  
		self._createQuery(poType) 
		query=poType.query
		if len (joinArr)>0:
			query.options(subqueryload(*joinArr)) 
		if  len (filters)>0:			
			query.filter(and_(*filters))
		if param !=None: 
			limit=param.limit
			offset=param.offset
			query.limit(limit).offset(offset)
		return (query.all()) 
	async def query_eager(self,poType:TypeVar, joinArr:Tuple[InstrumentedAttribute], *filters:ColumnElement[bool],param:Page_param=None,orderby: List[InstrumentedAttribute]=None):
		"""
		数据做笛卡儿积 ，程序为每个元素间应用连接条件进行解析
  		不应该使用
		"""  
		self._createQuery(poType) 
		query=poType.query
		if len (joinArr)>0:
			query.options(contains_eager(*joinArr)) 
		if  len (filters)>0:			
			query.filter(and_(*filters))
		if param !=None: 
			limit=param.limit
			offset=param.offset
			query.limit(limit).offset(offset)
		return (query.all()) 
	async def query_join_eager(self,poType:TypeVar, joinArr:Tuple[InstrumentedAttribute], *filters:ColumnElement[bool],param:Page_param=None,orderby: List[InstrumentedAttribute]=None):
		"""
		内连接，删除空
  		不应该使用
		"""  
		self._createQuery(poType) 
		query=poType.query
		if len (joinArr)>0:
			query.join(*joinArr)
			query.options(contains_eager(*joinArr)) 
		if  len (filters)>0:			
			query.filter(and_(*filters))
		if param !=None: 
			limit=param.limit
			offset=param.offset
			query.limit(limit).offset(offset)
		return (query.all()) 
	
class DbPagedOperations(DbOperations): 
	def __init__(self, db_session:AsyncSession,filter_items:absFilterItems):
		super().__init__(db_session)
		self.filter_items=filter_items  
	
	async def get_count(self,field:InstrumentedAttribute="*" )->int:
		"""
		根据 absFilterItems 获取符合条件的条数
		"""
		filters=self.filter_items.filter()
		return await self.count(*filters,field)
		#return await self._get_scalar(select(func.count(field)).filter(and_(*filters)))
		'''
		execute =await self.db_session.execute(select(func.count(field)).filter(and_(*filters)))  
		total= execute.scalar()   '''
		
		'''from co6co.utils import log
		#当选取整个对象的时候，都要用 scalars 方法，否则返回的是一个包含一个对象的 tuple 
		bts=await self.db_session.execute(select(ProcessPO.id,ProcessPO.boatName,ProcessPO.flowStatus).filter(and_(*filters)).limit(3))
	 
		log.succ(f"-----{type(bts)}\n{dir(bts)}") 
		for b in bts.fetchall():
			print(b)  

		filters=self.filter_items.filter( )
		orderBy=self.filter_items.getOrderBy() 
		data=await self.db_session.execute(select(ProcessPO).filter(and_(*filters)).offset(self.filter_items. get_db_page_index()).limit(3).order_by(*orderBy))
		log.succ(f"-----{type(data)}\n{dir(data)}")
		for b in data:
			print(b.tuple)  

		one=await self.db_session.get_one(ProcessPO,ident=2884) 
		print(one) ''' 
		return total
	
	async def get_paged (self ,selectColumnOrPo:Tuple[InstrumentedAttribute]|TypeVar=None,remove_instance_state:bool=True)-> List[dict]:
		"""
		selectColumn:  实体对象或者 filed
		返回列表
		"""
		isTule,sml=await self._create_paged_select(self.filter_items,selectColumnOrPo)
		if isTule:return await  self._get_tuple(sml)
		return await self._get_list(sml,remove_instance_state)




from typing import TypeVar, Tuple, List, Dict, Any, Union, Iterator, Callable
from sqlalchemy.engine.row import Row, RowMapping
from .po import BasePO
from co6co.utils import log

from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy import Select, Update, Insert, Delete
from sqlalchemy.future import select
from sqlalchemy import and_
from sqlalchemy import func, text
from sqlalchemy.sql.elements import ColumnElement
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.engine.result import ChunkedIteratorResult
from sqlalchemy.engine.cursor import CursorResult

from .db_filter import absFilterItems


class db_tools:
    """
    数据转换工具
    1. 
    data=  exec.mappings().all() 
    result=[dict(zip( a.keys(),a._to_tuple_instance())) for a in  data] 

    2.
    [dict(zip(a._fields,a))  for a in  executeResult]
    """
    __po_has_field__: str = "_sa_instance_state"

    @staticmethod
    def remove_db_instance_state(poInstance_or_poList: Iterator | Any) -> List[Dict] | Dict:
        if hasattr(poInstance_or_poList, "__iter__"):
            result = [dict(filter(lambda k: k[0] != db_tools.__po_has_field__,
                           a1.__dict__.items())) for a1 in poInstance_or_poList]
            for r in result:
                for r1 in r:
                    value = r.get(r1)
                    if (isinstance(value, BasePO)):
                        dic = db_tools.remove_db_instance_state(value)
                        r.update({r1: dic})
            return result
        # and hasattr (poInstance_or_poList,"__dict__")
        elif hasattr(poInstance_or_poList, "__dict__"):
            return dict(filter(lambda k: k[0] != db_tools.__po_has_field__, poInstance_or_poList.__dict__.items()))
        else:
            return poInstance_or_poList

    @staticmethod
    def row2dict(row: Row) -> Dict:
        """
        xxxxPO.id.label("xxx_id") 为数据取别名
        出现重名覆盖
        """
        d: dict = {}
        for i in range(0, len(row)):
            c = row[i]
            if hasattr(c, db_tools.__po_has_field__):
                dc = db_tools.remove_db_instance_state(c)
                d.update(dc)
            else:
                key = row._fields[i]
                '''
                j=1 
                while key in d.keys():
                    key=f"{ row._fields[i]}_{str(j )}"
                    j+=1
                '''
                d.update({key: c})
        return d

    @staticmethod
    def one2Dict(fetchone: Row | RowMapping) -> Dict:
        """
        Row:        execute.fetchmany() | execute.fetchone()
        RowMapping: execute.mappings().fetchall()|execute.mappings().fetchone()  
        """
        if type(fetchone) == Row:
            return dict(zip(fetchone._fields, fetchone))
        elif type(fetchone) == RowMapping:
            return dict(fetchone)
        elif type(fetchone) == dict:
            return fetchone
        log.warn(f"未知类型：‘{type(fetchone)}’,直接返回")
        return fetchone

    @staticmethod
    def list2Dict(list: List[Row | RowMapping]) -> List[dict]:
        return [db_tools.one2Dict(a) for a in list]

    def mapping(executeResult: any) -> List[dict]:
        """
        不在使用
        """
        # sqlalchemy.engine.result.ChunkedIteratorResult
        return [dict(zip(a._fields, a)) for a in executeResult]

    async def execSelect(session: AsyncSession, select: Select, params: Dict | List | Tuple = None) -> int | None:
        """
        执行查询语句
        @return: int | None
        """
        exec = await session.execute(select, params)
        return exec.scalar()

    async def count(session: AsyncSession, *filters: ColumnElement[bool], column: InstrumentedAttribute = "*") -> int:
        """
        count
        """
        sql = select(func.count(column)).filter(and_(*filters))
        return await db_tools.execSelect(session, sql)

    async def exist(session: AsyncSession, *filters: ColumnElement[bool], column: InstrumentedAttribute = "*") -> bool:
        """
        exist
        """
        count = await db_tools.count(session, *filters, column=column)
        if count > 0:
            return True
        else:
            return False

    async def execForMappings(session: AsyncSession, select: Select, queryOne: bool = False, params: Dict | Tuple | List = None):
        """
        session: AsyncSession
        select:Select 

        return list
        """

        executer = await session.execute(select, params)
        if queryOne:
            result = executer.mappings().fetchone()
            return db_tools.one2Dict(result)
        else:
            result = executer.mappings().all()
            return db_tools.list2Dict(result)

    async def execForPos(session: AsyncSession, select: Select, remove_db_instance_state: bool = True, params: Dict | List | Tuple = None):
        """
        session: AsyncSession
        select:Select
        remove_db_instance_state: bool

        return list
        """

        exec: ChunkedIteratorResult = await session.execute(select, params)
        if remove_db_instance_state:
            return db_tools.remove_db_instance_state(exec.scalars().fetchall())
        else:
            return exec.scalars().fetchall()

    async def execForPo(session: AsyncSession, select: Select, remove_db_instance_state: bool = True, params: Dict | List | Tuple = None):
        """
        session: AsyncSession
        select:Select
        remove_db_instance_state: bool

        return PO|None
        """
        exec: ChunkedIteratorResult = await session.execute(select, params)
        # user: UserPO = result.scalar()
        data = exec.fetchone()
        # 返回的是元组
        one = None
        if data != None:
            one = data[0]
            if remove_db_instance_state:
                return db_tools.remove_db_instance_state(one)
            else:
                return one
        else:
            return None

    async def execSQL(session: AsyncSession, sql: Update | Insert | Delete, sqlParam: Dict | List | Tuple = None) -> int:
        """
        执行简单SQL语句
        """
        data: CursorResult = await session.execute(sql, sqlParam)
        return data.rowcount


'''
exec.fetchone() //None| (data,)
exec.mappings().fetchone()  // {'id': 1, 'userName': 'admin'} | {"userPO":PO}
exec..fetchone()    //(1, 'admin') || po
'''


class DbCallable:
    session: AsyncSession = None

    def __init__(self, session: AsyncSession):
        self.session = session

    async def __call__(self, func: Callable[[AsyncSession], Any]):
        async with self.session, self.session.begin():
            if func != None:
                return await func(self.session)


class QueryOneCallable(DbCallable):
    async def __call__(self, select: Select, isPO: bool = True, param: Dict | List | Tuple = None):
        async def exec(session: AsyncSession):
            exec = await session.execute(select, param)
            if isPO:
                data = exec.fetchone()
                # 返回的是元组
                if data != None:
                    return data[0]
                else:
                    return None
            else:
                data = exec.mappings().fetchone()
                if data == None:
                    return None
                result = db_tools.one2Dict(data)
                return result
        return await super().__call__(exec)


class InsertCallable(DbCallable):
    async def __call__(self, *po: BasePO):
        async def exec(session: AsyncSession):
            try:
                session.add_all(po)
            except Exception as e:
                await session.rollback()
                log.warn("执行'InsertOneCallable'异常", e)
                raise
        return await super().__call__(exec)


class UpdateOneCallable(DbCallable):
    async def __call__(self, queryOneSelect: Select, editFn: Callable[[AsyncSession, Any], None | Any] = None, param: Dict | List | Tuple = None):
        """
        queryOneSelect: 查询语句
        editFn: (session,po)->Any|None   返回:None  ->回滚,
                                            :Any   -> 函数返回值
        """
        async def exec(session: AsyncSession):
            try:
                exec = await session.execute(queryOneSelect, param)
                data = exec.fetchone()
                # 返回的是元组
                one = None
                if data != None:
                    one = data[0]
                else:
                    one = None
                if editFn != None:
                    result = await editFn(session, one)
                    if result == None:
                        await session.rollback()
                    return result
            except Exception as e:
                await session.rollback()
                log.warn("执行'UpdateOneCallable'异常", e)
                raise

        return await super().__call__(exec)


class QueryListCallable(DbCallable):
    async def __call__(self, select: Select, isPO: bool = True, remove_db_instance=True, param: Dict | List | Tuple = None):
        async def exec(session: AsyncSession):
            if isPO:
                result = await db_tools.execForPos(session, select, remove_db_instance, params=param)
            else:
                result = await db_tools.execForMappings(session, select, params=param)
            return result
        # return await super(QueryListCallable,self).__call__(exec) #// 2.x 写法
        return await super().__call__(exec)


class QueryPagedCallable(DbCallable):
    async def __call__(self, countSelect: Select, select: Select, isPO: bool = True, remove_db_instance=True, param: Dict | List | Tuple = None) -> Tuple[int, List[dict]]:
        async def exec(session: AsyncSession):
            total = await db_tools.execSelect(session, countSelect, param)
            if isPO:
                result = await db_tools.execForPos(session, select, remove_db_instance, param)
            else:
                result = await db_tools.execForMappings(session, select, param)

            return total, result
        return await super().__call__(exec)


class QueryPagedByFilterCallable(QueryPagedCallable):
    async def __call__(self, filter: absFilterItems, isPO: bool = True, remove_db_instance=True, param: Dict | List | Tuple = None) -> Tuple[int, List[dict]]:
        return await super().__call__(filter.count_select, filter.list_select, isPO, remove_db_instance, param)



from sqlalchemy.ext.asyncio import create_async_engine, AsyncEngine

from contextvars import ContextVar

from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import scoped_session, sessionmaker

# from model.pos.DbModelPo import User, Process

from sqlalchemy import select, create_engine, Engine
from sqlalchemy.orm import selectinload

import asyncio
import time
import typing
from typing import TypeVar
from co6co.utils import log
from co6co_db_ext.po import BasePO
from sqlalchemy.pool import NullPool


class db_service:
    default_settings: dict = {
        'DB_HOST': 'localhost',
        'DB_NAME': '',
        'DB_USER': 'root',
        'DB_PASSWORD': '',
        'echo': True,
        'pool_size': 20,
        'max_overflow': 10,
        'pool_pre_ping': True,  # 执行sql语句前悲观地检查db是否可用
        # 'pool_recycle':1800 #超时时间 单位s

    }
    settings = {}
    session: scoped_session  # 同步连接

    async_session_factory: sessionmaker  # 异步连接
    """
    AsyncSession 工厂函数
    sessionmaker 是个生成器类

    """
    useAsync: bool
    poolSize: int = None
    poolSize: int = None

    def createEngine(self, url, **kwargs) -> Engine:
        setting = {
            "ping": self.settings.get("pool_pre_ping"),
            "echo": True if type(self.settings.get("echo")) != bool else self.settings.get("echo"),
            "pool_size": self.settings.get("pool_size"),
            "max_overflow": self.settings.get("max_overflow"),
            "poolclass": NullPool
        }
        setting.update(kwargs)
        return create_engine(url, **setting)

    def createAsyncEngine(self, url, **kwargs) -> AsyncEngine:
        setting = {
            "pool_pre_ping": self.settings.get("pool_pre_ping"),
            "echo": True if type(self.settings.get("echo")) != bool else self.settings.get("echo"),
            "pool_size": self.settings.get("pool_size"),
            "max_overflow": self.settings.get("max_overflow")
        }
        setting.update(kwargs)
        return create_async_engine(url, **setting)

    def _session_factory(self, engine: Engine = None, **kv) -> AsyncSession:
        if engine == None:
            engine = self.createEngine(self.url)
        default = {
            "autoflush": False,
            "autocommit": False,
        }
        default.update(kv)
        factory = sessionmaker(bind=self.engine,  **default)
        return factory

    def _async_session_factory(self, engine: AsyncEngine = None, **kv) -> AsyncSession:
        if engine == None:
            engine = self.createAsyncEngine(self.url)
        default = {
            "expire_on_commit": False,
            "class_": AsyncSession,
        }
        default.update(kv)
        factory = sessionmaker(engine, **default)
        return factory

    def createSession(self, engine: AsyncEngine = None, **kv) -> scoped_session:
        factory = self._session_factory(engine, **kv)
        return scoped_session(factory)

    def createAsyncSession(self, engine: AsyncEngine = None, **kv) -> AsyncSession:
        factory = self._async_session_factory(engine, **kv)
        return factory()

    def _createEngine(self, url: str):
        self.useAsync = True
        if "sqlite" in url:
            self.useAsync = False
            self.engine = self.createEngine(url)
            self.session = self.createSession(self.engine)
            BasePO.query = self.session.query_property()
        else:  # AsyncSession
            self.engine = self.createAsyncEngine(url)
            self.async_session_factory = self._async_session_factory(self.engine)

        self.base_model_session_ctx = ContextVar("session")
        pass

    def __init__(self, config: dict, engineUrl: str = None) -> None:
        self.settings = self.default_settings.copy()
        if engineUrl == None:
            self.settings .update(config)
            engineUrl = "mysql+aiomysql://{}:{}@{}/{}".format(self.settings['DB_USER'], self.settings['DB_PASSWORD'], self.settings['DB_HOST'], self.settings['DB_NAME'])
        self.url = engineUrl
        self._createEngine(engineUrl)
        pass

    async def init_tables(self):
        if self.useAsync:
            async with self.engine.begin() as conn:
                # await conn.run_sync(BasePO.metadata.drop_a顶顶顶顶ll)
                await conn.run_sync(BasePO.metadata.create_all)
                await conn.commit()
            await self.engine.dispose()
        else:
            BasePO.metadata.create_all(bind=self.engine)

    def sync_init_tables(self):
        retryTime = 0
        while (True):
            try:
                if retryTime < 8:
                    retryTime += 1
                asyncio.run(self.init_tables())
                break
            except Exception as e:
                log.warn(f"同步数据表失败{e}!")
                log.info(f"{retryTime*5}s后重试...")
                time.sleep(retryTime*5)



import asyncio
from co6co.utils import log
from sqlalchemy.ext.asyncio import AsyncSession, AsyncConnection, AsyncEngine
from co6co_db_ext.db_session import db_service
from typing import List
import asyncio
from sanic import Sanic
from co6co.task.thread import ThreadEvent


class DbSession:
    """
    创建 Session 
    """
    session: AsyncSession = None
    engine: AsyncEngine = None

    def __init__(self, db_settings: dict) -> None:
        _service = db_service(db_settings)
        self.session: AsyncSession = _service.async_session_factory()
        self.engine = _service.engine
        '''
        service:db_service=app.ctx.service
        self.session:AsyncSession=service.async_session_factory()
        '''
        # log.warn(f"..创建session。。")
        pass

    def __del__(self) -> None:
        if self.session:
            asyncio.run(self.session.close)
        # log.info(f"{self}...关闭session")

    def __repr__(self) -> str:
        return f'{self.__class__}'


class BaseBll(DbSession):
    """
    数据库操作
    定义异步方法
    运行 result=run(异步方法,arg) 
    """

    _eventLoop: ThreadEvent

    def __init__(self, db_settings: dict) -> None:
        self._eventLoop = ThreadEvent()
        super().__init__(db_settings)

    def run(self, task, *args, **argkv):
        data = self._eventLoop.runTask(task, *args, **argkv)
        return data

    def __del__(self) -> None:
        if self.session:
            self.run(self.session.close)
            '''
            conn: AsyncConnection = self.run(self.session.connection) 
            self.run( conn.commit)
            log.warn("关闭连接",conn._proxied.__hash__())
            if not conn.closed: 
                self.run(conn.close)
            log.warn("关闭连接{}", conn.closed)
            '''
            # 不关闭 engine 会报 connect 关闭异常的错误
            # 就算关闭连接也不行
            self.run(self.engine.dispose)

        self._eventLoop.close()
        # if self.session: await self.session.close()




from abc import ABC, abstractmethod 
from .page_param import Page_param
from typing import TypeVar,Tuple,List,Dict,Any,Union,Iterator
from sqlalchemy .orm.attributes import InstrumentedAttribute
from sqlalchemy.sql.elements import ColumnElement
from sqlalchemy import func,or_,and_,Select  

class absFilterItems(ABC, Page_param): 
	"""
	抽象过滤器
	配合 DbOperations 使用
	"""
	listSelectFields: List[InstrumentedAttribute]=None# 使用该数据 dictor
	po_type:TypeVar=None
	def __init__(self,po:TypeVar) -> None:
		super().__init__() 

		self.po_type=po # 与排序相关
		 
		pass

	@property
	def offset(self):
		return self.get_db_page_index()*self.pageSize
	@property
	def limit(self):
		return self.pageSize
	

	#@abstractclassmethod 
	@abstractmethod
	def filter(self)->List[ColumnElement[bool]]:
		raise NotADirectoryError("Can't instantiate abstract clas") 
	
	def _getOrderby(self)->List[Dict[str,str]]:
		"""
		orderBy:id,name
		order: asc,desc
		从字符串转 [{"id":"asc"},{"name","desc"}]
		"""
		if self.orderBy and  self.order:
			by=self.orderBy.split(",")
			order=self.order.split(",")
			if len(by)==len(order):
				return [{b:o} for b in by for o in order if b and o]
			else:
				return [{b:order[0]} for b in by if b ]
		elif self.orderBy:
			by=self.orderBy.split(",")
			return [{b:"asc"} for b in by if b ]
		return [] 
	 
	@abstractmethod
	def getDefaultOrderBy(self)->Tuple[InstrumentedAttribute]:
		raise NotADirectoryError("Can't instantiate abstract clas") 
	 
	def getOrderBy(self)->List[InstrumentedAttribute]:
		""" 
		获取排序规则
		
		"""
		orderList= self._getOrderby()
		if len(orderList)==0: return self.getDefaultOrderBy()
		# () 获取的结果不能重复 取*
        # []  获取的结果能重复 取* 
		else: return [ self. po_type.__dict__[key].desc() if it[key] and it[key].lower()=="desc" else  self. po_type.__dict__[key].asc() for it in orderList for key in it.keys() if key in self.po_type.__dict__]
	def getOrderBy_Fields(self)->List[InstrumentedAttribute]:
		""" 
		获取排序规则 2 由 select 字段 排序
		//todo 功能未实现
		"""
		orderList= self._getOrderby()
		if len(orderList)==0: return self.getDefaultOrderBy() 
		else: return [ self. po_type.__dict__[key].desc() if it[key] and it[key].lower()=="desc" else  self. po_type.__dict__[key].asc() for it in orderList for key in it.keys() if key in self.po_type.__dict__]
	
	def checkFieldValue(self,fielValue:Any)->bool:
		if type(fielValue) == str and fielValue:return True
		if type(fielValue) == int :return True
		if type(fielValue) == bool :return True
		return False
	 
	def create_List_select(self):
		if self.listSelectFields !=None and len(self.listSelectFields)>0: 
			select=(
				Select(*self.listSelectFields)#.join(device.deviceCategoryPO,isouter=True)
				.filter(and_(*self.filter()))  
			)
		else :
			select=(
				Select(self.po_type)#.join(device.deviceCategoryPO,isouter=True)
				.filter(and_(*self.filter()))  
			) 
		return select
	
	@property
	def list_select(self,useListFiled=False):
		if useListFiled:
			return self.create_List_select().limit(self.limit).offset(self.offset).order_by(*self.getOrderBy_Fields())
		return self.create_List_select().limit(self.limit).offset(self.offset).order_by(*self.getOrderBy())
	@property
	def count_select(self): 
		return Select( func.count( )).select_from(self.create_List_select())



from setuptools import setup, find_packages

import co6co_db_ext
VERSION = co6co_db_ext.__version__

# read readmeFile contents
from os import path
currentDir = path.abspath(path.dirname(__file__))
with open(path.join(currentDir, 'README.md'), encoding='utf-8') as f: long_description = f.read()


setup(
    name="co6co.db_ext",
    version=VERSION,
    description="db 数据库扩展",
    packages=find_packages(),
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[ "Programming Language :: Python :: 3", "Programming Language :: Python :: 3.6" ],
    include_package_data=True, zip_safe=True,
    #依赖哪些模块
    install_requires=["co6co","aiomysql","SQLAlchemy" ],
    #package_dir= {'utils':'src/log','main_package':'main'},#告诉Distutils哪些目录下的文件被映射到哪个源码
    author='co6co',
    author_email ='co6co@qq.com',
    url="http://github.com/co6co",
    data_file={
        ('',"*.txt"),
        ('',"*.md"),
    },
    package_data={
        '':['*.txt','*.md'],
        'bandwidth_reporter':['*.txt']
    }
)


from __future__ import annotations
from math import ceil


class Page_param:
    def __init__(self) -> None:
        self.pageIndex = 1
        self.pageSize = 10
        self.orderBy = ""
        self.order = "asc"  # [desc|asc]
        pass

    def get_db_page_index(self):
        return self.pageIndex-1

    def getMaxPageIndex(self, recodeCount: int):
        return ceil(recodeCount/self.pageSize)



from __future__ import annotations
from sqlalchemy import func, MetaData, INTEGER, Column, ForeignKey, String, BigInteger, DateTime
from sqlalchemy.orm import DeclarativeBase, declarative_base, relationship, QueryPropertyDescriptor, Query
from co6co.utils.tool_util import to_camelcase
from datetime import datetime
# 与属性库实体相关，所有实体集成 BasePO类

metadata = MetaData()
# class BasePO(declarative_base(metadata=metadata)):


class BasePO(DeclarativeBase):
    __abstract__ = True

    @property
    def query() -> Query:
        return None

    def to_dict(self):
        return dict(filter(lambda k: k[0] != "_sa_instance_state", self.__dict__.items()))
        # return {to_camelcase(c.name): getattr(self, c.name, None) for c in self.__table__.columns}

    def to_dict2(self):
        return {to_camelcase(c.name): getattr(self, to_camelcase(c.name), None) for c in self.__table__.columns}

    def update(self, po: DeclarativeBase):
        pass

    def add_assignment(self,  userId: None):
        if isinstance(self, UserTimeStampedModelPO):
            self.createTime = datetime.now()
            self.createUser = userId
        elif isinstance(self, TimeStampedModelPO):
            self.createTime = datetime.now()

    def edit_assignment(self,   userId: None):
        if isinstance(self, UserTimeStampedModelPO):
            self.updateTime = datetime.now()
            self.updateUser = userId
        elif isinstance(self, TimeStampedModelPO):
            self.updateTime = datetime.now()


class TimeStampedModelPO(BasePO):
    __abstract__ = True

    createTime = Column("create_time", DateTime, server_default=func.now(), comment="创建时间")
    updateTime = Column("update_time", DateTime, comment="更新时间")


class CreateUserStampedModelPO(BasePO):
    __abstract__ = True
    createUser = Column("create_user", BigInteger, comment="创建人")
    createTime = Column("create_time", DateTime, server_default=func.now(), comment="创建时间")


class UserTimeStampedModelPO(TimeStampedModelPO):
    __abstract__ = True

    createUser = Column("create_user", BigInteger, comment="创建人")
    updateUser = Column("update_user", BigInteger, comment="修改人")
